class ItemsCursor(object):

    def __init__(self,
                 context,
                 container_name,
                 access_key,
                 path,
                 raise_for_status=None,
                 table_name=None,
                 attribute_names='*',
                 filter_expression=None,
                 marker=None,
                 sharding_key=None,
                 limit=None,
                 segment=None,
                 total_segments=None,
                 sort_key_range_start=None,
                 sort_key_range_end=None):
        self._context = context
        self._container_name = container_name
        self._access_key = access_key
        self._current_response = None
        self._current_items = None
        self._current_item = None
        self._current_item_index = 0

        # get items params
        self.raise_for_status = raise_for_status
        self.path = path
        self.table_name = table_name
        self.attribute_names = attribute_names
        self.filter_expression = filter_expression
        self.marker = marker
        self.sharding_key = sharding_key
        self.limit = limit
        self.segment = segment
        self.total_segments = total_segments
        self.sort_key_range_start = sort_key_range_start
        self.sort_key_range_end = sort_key_range_end

    def next_item(self):
        if self._current_item_index < len(self._current_items or []):
            self._current_item = self._current_items[self._current_item_index]
            self._current_item_index += 1

            return self._current_item

        if self._current_response and self._current_response.output.last:
            return None

        self.marker = self._current_response.output.next_marker if self._current_response else None

        # get the next batch
        self._current_response = self._context.get_items(self._container_name,
                                                         self.path,
                                                         self._access_key,
                                                         self.raise_for_status,
                                                         None,
                                                         self.table_name,
                                                         self.attribute_names,
                                                         self.filter_expression,
                                                         self.marker,
                                                         self.sharding_key,
                                                         self.limit,
                                                         self.segment,
                                                         self.total_segments,
                                                         self.sort_key_range_start,
                                                         self.sort_key_range_end)

        # raise if there was an issue
        self._current_response.raise_for_status()

        # set items
        self._current_items = self._current_response.output.items
        self._current_item_index = 0

        # and recurse into next now that we repopulated response
        return self.next_item()

    def all(self):
        items = []

        while True:
            item = self.next_item()

            if item is None:
                break

            items.append(item)

        return items



import ssl
import http.client

import v3io.dataplane.response
import v3io.dataplane.request
from . import abstract


class Transport(abstract.Transport):

    def __init__(self, logger, endpoint=None, max_connections=None, timeout=None):
        super(Transport, self).__init__(logger, endpoint, max_connections, timeout)

        # holds which connection index we'll use
        self._next_connection_idx = 0

        # based on scheme, create a host and context for _create_connection
        self._host, self._ssl_context = self._parse_endpoint(self._endpoint)

        # create the pool connection
        self._connections = self._create_connections(self.max_connections,
                                                     self._host,
                                                     self._ssl_context)

    def restart(self):
        self.close()

        # recreate the connections
        self._connections = self._create_connections(self.max_connections,
                                                     self._host,
                                                     self._ssl_context)

    def close(self):
        for connection in self._connections:
            connection.close()

    def send_request(self, request, transport_state=None):
        if transport_state is None:
            connection_idx = self._get_next_connection_idx()
        else:
            connection_idx = transport_state.connection_idx

        # set the used connection on the request
        setattr(request.transport, 'connection_idx', connection_idx)

        # get a connection for the request and send it
        return self._send_request_on_connection(request, connection_idx)

    def wait_response(self, request, raise_for_status=None, num_retries=1):
        connection_idx = request.transport.connection_idx
        connection = self._connections[connection_idx]

        while True:
            try:

                # read the response
                response = connection.getresponse()
                response_body = response.read()

                response = v3io.dataplane.response.Response(request.output,
                                                            response.code,
                                                            response.headers,
                                                            response_body)

                # enforce raise for status
                response.raise_for_status(request.raise_for_status or raise_for_status)

                # return the response
                return response

            except http.client.RemoteDisconnected as e:
                if num_retries == 0:
                    raise e

                num_retries -= 1

                # create a connection
                connection = self._recreate_connection_at_index(connection_idx)

                # re-send the request on the connection
                request = self._send_request_on_connection(request, connection_idx)

    def _send_request_on_connection(self, request, connection_idx):
        connection = self._connections[connection_idx]

        try:
            connection.request(request.method, request.path, request.body, request.headers)
        except (BrokenPipeError, http.client.CannotSendRequest, http.client.RemoteDisconnected):
            connection = self._recreate_connection_at_index(connection_idx)

            # re-request
            connection.request(request.method, request.path, request.body, request.headers)

        return request

    def _recreate_connection_at_index(self, connection_idx):

        # close the old connection
        self._connections[connection_idx].close()

        # create a new one and replace
        connection = self._create_connection(self._host, self._ssl_context)
        self._connections[connection_idx] = connection

        return connection

    def _create_connections(self, num_connections, host, ssl_context):
        connections = []

        for connection_idx in range(num_connections):
            connection = self._create_connection(host, ssl_context)
            connection.connect()
            connections.append(connection)

        return connections

    def _create_connection(self, host, ssl_context):
        if ssl_context is None:
            return http.client.HTTPConnection(host)

        return http.client.HTTPSConnection(host, context=ssl_context)

    def _parse_endpoint(self, endpoint):
        if endpoint.startswith('http://'):
            return endpoint[len('http://'):], None

        if endpoint.startswith('https://'):
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

            return endpoint[len('https://'):], ssl_context

        return endpoint, None

    def _get_next_connection_idx(self):
        connection_idx = self._next_connection_idx

        self._next_connection_idx += 1
        if self._next_connection_idx >= len(self._connections):
            self._next_connection_idx = 0

        return connection_idx



import ujson
import xml.etree.ElementTree

import v3io.dataplane.transport


class Response(object):

    def __init__(self, output, status_code, headers, body):
        self.status_code = status_code
        self.body = body
        self.headers = headers
        self.output = None

        if output and self.body:
            try:
                parsed_output = ujson.loads(self.body)
            except Exception:
                parsed_output = xml.etree.ElementTree.fromstring(self.body)

            self.output = output(parsed_output)

    def raise_for_status(self, expected_statuses=None):
        if expected_statuses == v3io.dataplane.transport.RaiseForStatus.never:
            return

        # "always" and "none" are equivalent. use the one that's faster to compare against
        if expected_statuses == v3io.dataplane.transport.RaiseForStatus.always:
            expected_statuses = None

        if (expected_statuses is None and self.status_code >= 300) \
                or (expected_statuses and self.status_code not in expected_statuses):
            raise RuntimeError('Request failed with status {0}: {1}'.format(self.status_code, self.body))


class Responses(object):

    def __init__(self):
        self.responses = []
        self.success = True

    def add_response(self, response):
        self.responses.append(response)

        if response.status_code != 200:
            self.success = False

    def raise_for_status(self):
        if not self.success:
            raise RuntimeError('Failed to put items')
